searchData={"items":[{"type":"module","title":"Grammar","doc":"This module provides a DSL to define parser of structured inputs. Parsers are defined as a grammar.\n\nGrammar defined must be LL(1) grammars, i.e. they must be unambiguous and have a single token lookahead.\n\nThe grammar is defined by a set of rules, each rule being a set of clauses. Clauses must be understood as\ndisjoinded paths in the rule resolution, or as the `or` operator in the classic notation.\n\nThe tokenization process relies on the [TokenExtractor](`Grammar.Tokenizer.TokenExtractor`) protocol, which is used to extract tokens from the input string.\nThis protocol is implemented for BitString and Regex, and can be extended to custom token types.\n\nTo declare a parser module, just `use` the Grammar module in your module, and define your rules using the `rule/2` and `rule?/2` macro.\nThe newly defined module will expose a `parse/1` function that will parse the input string, and return a tuple with the\ntokenizer in its final state, and result.\n\nSee `rule/2` for a full example.","ref":"Grammar.html"},{"type":"module","title":"Spaces and line breaks handling - Grammar","doc":"By default, the tokenizer will drop spaces and line breaks.\nIf you want to keep them, you can pass the `drop_spaces: false` option to the `use Grammar` macro. In this case, you\nare fully responsible for handling spaces and line breaks in your rules.","ref":"Grammar.html#module-spaces-and-line-breaks-handling"},{"type":"module","title":"Options - Grammar","doc":"- `drop_spaces: true` (default): if set to `false`, the tokenizer will not drop spaces and line breaks.\n\n#### Example\n\nIn the following `MyModuleKO` module, the `start` rule doesn't handle spaces and line breaks,\nso it will fail if the input contains them.\n\n    iex> defmodule MyModuleKO do\n    ...>   use Grammar, drop_spaces: false\n    ...>\n    ...>   # spaces and linebreaks not handled\n    ...>\n    ...>   rule start(\"hello\", \"world\") do\n    ...>     [_hello, _world] = params\n    ...>     \"hello world\"\n    ...>   end\n    ...> end\n    iex> MyModuleKO.parse(\"helloworld\")\n    {:ok, \"hello world\"}\n    iex> MyModuleKO.parse(\"hello world\")\n    {:error, {1, 6}, :no_token}\n\nBut in the `MyModuleOK` module, the `start` explicitly handles spaces and line breaks between \"hello\" and \"world\".\n\nAnd even more that rule definition requires at least one space between \"hello\" and \"world\", parsing fails if no space is found.\n\n    iex> defmodule MyModuleOK do\n    ...>   use Grammar, drop_spaces: false\n    ...>\n    ...>   # spaces and linebreaks not handled\n    ...>\n    ...>   rule start(\"hello\", ~r/[\\s]+/, \"world\") do\n    ...>     [_hello, _spaces, _world] = params\n    ...>     \"hello world\"\n    ...>   end\n    ...> end\n    iex> MyModuleOK.parse(\"helloworld\")\n    {:error, {1, 6}, :no_token}\n    iex> MyModuleOK.parse(~s/hello  \\t world/)\n    {:ok, \"hello world\"}","ref":"Grammar.html#module-options"},{"type":"macro","title":"Grammar.rule/2","doc":"Use this macro to define rules of your grammar.\n\n**The first rule defined will be the entry rule of the grammar**.\n\nCalls to this macro sharing the same name will be grouped together as they define the same rule,\neach call is a possible path in the rule resolution.\n\nLets name a single call to `rule` a clause.\nAll clauses must be disjointed, i.e. they must not share the same first token.\nThey can be understood as the `or` operator in a rule.\n\nEach rule of rule clause is defined by\n- a name, which is an atom\n- a definition, which is a list of atoms or token prototypes\n- a block, which is the code to execute when the clause is fully matched\n\nWhen executed the code block is provided with a `params` binding, which is a list of the results of the clause steps.\n\nIn the case where a `rule` cannot be matched, a `RuntimeError` is raised (see `rule?/2` for a relaxed version).","ref":"Grammar.html#rule/2"},{"type":"macro","title":"Example - Grammar.rule/2","doc":"iex> defmodule NumberOfNameListParser do\n    ...>   use Grammar\n    ...>\n    ...>   rule start(\"[\", :list_or_empty_list) do\n    ...>     [_, list] = params\n    ...>     list || []\n    ...>   end\n    ...>\n    ...>   rule? list_or_empty_list(:item, :list_tail, \"]\") do\n    ...>     [item, list_tail, _] = params\n    ...>     [item | (list_tail || [])]\n    ...>   end\n    ...>\n    ...>   rule? list_tail(\",\", :item, :list_tail) do\n    ...>     [_, item, list_tail] = params\n    ...>     [item | (list_tail || [])]\n    ...>   end\n    ...>\n    ...>   rule item(~r/[0-9]+/) do\n    ...>     [number] = params\n    ...>     String.to_integer(number)\n    ...>   end\n    ...>\n    ...>   rule item(~r/[a-zA-Z]+/) do\n    ...>     [string] = params\n    ...>     string\n    ...>   end\n    ...> end\n    iex> GrammarTest.NumberOfNameListParser.parse(\"[1, toto, 23]\")\n    {:ok, [1, \"toto\", 23]}","ref":"Grammar.html#rule/2-example"},{"type":"macro","title":"Grammar.rule?/2","doc":"Same as `rule/2` but relaxed : if the rule cannot be matched, it will be valued as `nil`.\n\nUseful for optional or recursive rules.\n\nSee example in `rule/2`.","ref":"Grammar.html#rule?/2"},{"type":"protocol","title":"Grammar.Tokenizer.TokenExtractor","doc":"This protocol exposes the functions needed to extract tokens from the input string.","ref":"Grammar.Tokenizer.TokenExtractor.html"},{"type":"function","title":"Grammar.Tokenizer.TokenExtractor.match?/2","doc":"Returns true if the token matches the token prototype.\n\nThis function is used orient the parser in the right clause of a rule, by comparing the current token\nto the clause list of first tokens.","ref":"Grammar.Tokenizer.TokenExtractor.html#match?/2"},{"type":"function","title":"Example - Grammar.Tokenizer.TokenExtractor.match?/2","doc":"iex> Grammar.Tokenizer.TokenExtractor.match?(\"hello\", \"hello\")\n    true\n\n    iex> Grammar.Tokenizer.TokenExtractor.match?(\"hello\", \"Horld\")\n    false\n\n    iex> Grammar.Tokenizer.TokenExtractor.match?(~r/[0-9]+/, \"013456\")\n    true\n\n    iex> Grammar.Tokenizer.TokenExtractor.match?(~r/[0-9]+/, \"a013456\")\n    false","ref":"Grammar.Tokenizer.TokenExtractor.html#match?/2-example"},{"type":"function","title":"Grammar.Tokenizer.TokenExtractor.try_read/2","doc":"Try to read a token from the input string.\n\nIf the token is found, returns the token and its length in the input string.\nReturns nil otherwise.","ref":"Grammar.Tokenizer.TokenExtractor.html#try_read/2"},{"type":"function","title":"Example - Grammar.Tokenizer.TokenExtractor.try_read/2","doc":"iex> Grammar.Tokenizer.TokenExtractor.try_read(\"hello\", \"hello world\")\n    {\"hello\", 5}\n\n    iex> Grammar.Tokenizer.TokenExtractor.try_read(\"Hello\", \"hello world\")\n    nil\n\n    iex> Grammar.Tokenizer.TokenExtractor.try_read(~r/[0-9]+/, \"013456toto\")\n    {\"013456\", 6}\n\n    iex> Grammar.Tokenizer.TokenExtractor.try_read(~r/[a-z]+/, \"013456toto\")\n    nil","ref":"Grammar.Tokenizer.TokenExtractor.html#try_read/2-example"},{"type":"type","title":"Grammar.Tokenizer.TokenExtractor.t/0","doc":"All the types that implement this protocol.","ref":"Grammar.Tokenizer.TokenExtractor.html#t:t/0"},{"type":"module","title":"Grammar.Tokenizer.TokenExtractorHelper","doc":"This module provides helper functions to work with TokenExtractor implementations using Regex.","ref":"Grammar.Tokenizer.TokenExtractorHelper.html"},{"type":"function","title":"Grammar.Tokenizer.TokenExtractorHelper.normalize_regex/1","doc":"Normalizes a regex to ensure it starts with `^`, while keeping the original options.","ref":"Grammar.Tokenizer.TokenExtractorHelper.html#normalize_regex/1"},{"type":"function","title":"Grammar.Tokenizer.TokenExtractorHelper.try_read_from_regex/2","doc":"Tries to read a token from the input string using a regex pattern.","ref":"Grammar.Tokenizer.TokenExtractorHelper.html#try_read_from_regex/2"},{"type":"module","title":"Grammar.CodeGen","doc":"This module exposes functions required to generate the code for the parser derived from grammar rules.","ref":"Grammar.CodeGen.html"},{"type":"function","title":"Grammar.CodeGen.build_production_for_rules/1","doc":"","ref":"Grammar.CodeGen.html#build_production_for_rules/1"},{"type":"function","title":"Grammar.CodeGen.check_all_rules_exist/1","doc":"","ref":"Grammar.CodeGen.html#check_all_rules_exist/1"},{"type":"function","title":"Grammar.CodeGen.check_rules_are_not_ambiguous/1","doc":"","ref":"Grammar.CodeGen.html#check_rules_are_not_ambiguous/1"},{"type":"function","title":"Grammar.CodeGen.first_of_rules/1","doc":"","ref":"Grammar.CodeGen.html#first_of_rules/1"},{"type":"function","title":"Grammar.CodeGen.store_clause/6","doc":"","ref":"Grammar.CodeGen.html#store_clause/6"},{"type":"module","title":"Grammar.CodeGen.Clause","doc":"A Clause entity represents a grammar clause, by its definition and the block to execute.\n\nThe definition is a list of steps, which can be either a rule name or a token prototype.\n\nThe block is the code to execute when the clause is fully matched. In this code block,\nthe binding `params` is available and contains the results of the clause steps.","ref":"Grammar.CodeGen.Clause.html"},{"type":"function","title":"Grammar.CodeGen.Clause.epsilon?/1","doc":"","ref":"Grammar.CodeGen.Clause.html#epsilon?/1"},{"type":"function","title":"Grammar.CodeGen.Clause.new/3","doc":"","ref":"Grammar.CodeGen.Clause.html#new/3"},{"type":"function","title":"Grammar.CodeGen.Clause.set_firsts/2","doc":"","ref":"Grammar.CodeGen.Clause.html#set_firsts/2"},{"type":"module","title":"Grammar.CodeGen.Rule","doc":"A Rule entity represents a grammar rule, by its name and the list of clauses.","ref":"Grammar.CodeGen.Rule.html"},{"type":"function","title":"Grammar.CodeGen.Rule.add_clause/2","doc":"","ref":"Grammar.CodeGen.Rule.html#add_clause/2"},{"type":"function","title":"Grammar.CodeGen.Rule.add_clause/4","doc":"","ref":"Grammar.CodeGen.Rule.html#add_clause/4"},{"type":"function","title":"Grammar.CodeGen.Rule.get_firsts/1","doc":"","ref":"Grammar.CodeGen.Rule.html#get_firsts/1"},{"type":"function","title":"Grammar.CodeGen.Rule.new/1","doc":"","ref":"Grammar.CodeGen.Rule.html#new/1"},{"type":"function","title":"Grammar.CodeGen.Rule.new/4","doc":"","ref":"Grammar.CodeGen.Rule.html#new/4"},{"type":"module","title":"Grammar.Tokenizer","doc":"This module extracts the tokens from the input string.\n\nIt is driven by the parser to search for specific tokens only, when required.","ref":"Grammar.Tokenizer.html"},{"type":"function","title":"Grammar.Tokenizer.current_token/2","doc":"Returns the current token found in the input, if any.\nExpected tokens are passed as a list of token prototypes.\n\nThe token is not consumed, so two succesive calls to `current_token/2` will return the same token.","ref":"Grammar.Tokenizer.html#current_token/2"},{"type":"function","title":"Examples - Grammar.Tokenizer.current_token/2","doc":"iex> tokenizer = Grammar.Tokenizer.new(\"hello world\")\n    %Grammar.Tokenizer{\n      input: \"hello world\",\n      current_line: 1,\n      current_column: 1,\n      drop_spaces?: true\n    }\n    iex> {{\"hello\", {1, 1}}, _} = Grammar.Tokenizer.current_token(tokenizer, [\"hello\"])\n    iex> {{\"hello\", {1, 1}}, _} = Grammar.Tokenizer.current_token(tokenizer, [\"hello\"])\n    iex> {{nil, {1, 1}}, _} = Grammar.Tokenizer.current_token(tokenizer, [\"world\"])","ref":"Grammar.Tokenizer.html#current_token/2-examples"},{"type":"function","title":"Grammar.Tokenizer.new/2","doc":"Creates a new tokenizer for a given input.","ref":"Grammar.Tokenizer.html#new/2"},{"type":"function","title":"Parameters - Grammar.Tokenizer.new/2","doc":"- input: bitstring from which token will be extracted.\n  - drop_spaces?: when set to `false`, the tokenizer will not drop spaces and newlines.","ref":"Grammar.Tokenizer.html#new/2-parameters"},{"type":"function","title":"Examples - Grammar.Tokenizer.new/2","doc":"iex> Grammar.Tokenizer.new(\"this is the input\")\n    %Grammar.Tokenizer{\n      input: \"this is the input\",\n      current_line: 1,\n      current_column: 1,\n      drop_spaces?: true\n    }\n\n    iex> Grammar.Tokenizer.new(\"this is the input\", false)\n    %Grammar.Tokenizer{\n      input: \"this is the input\",\n      current_line: 1,\n      current_column: 1,\n      drop_spaces?: false\n    }","ref":"Grammar.Tokenizer.html#new/2-examples"},{"type":"function","title":"Grammar.Tokenizer.next_token/2","doc":"Returns the current token found in the input, and consumes it.\nThe expected token prototype is passed as second parameter.","ref":"Grammar.Tokenizer.html#next_token/2"},{"type":"function","title":"Examples - Grammar.Tokenizer.next_token/2","doc":"iex> tokenizer = Grammar.Tokenizer.new(\"hello world\")\n    iex> {{\"hello\", {1, 1}}, tokenizer} = Grammar.Tokenizer.next_token(tokenizer, \"hello\")\n    iex> {{\"world\", {1, 7}}, _} = Grammar.Tokenizer.next_token(tokenizer, \"world\")","ref":"Grammar.Tokenizer.html#next_token/2-examples"},{"type":"type","title":"Grammar.Tokenizer.t/0","doc":"","ref":"Grammar.Tokenizer.html#t:t/0"},{"type":"extras","title":"Grammar","doc":"# Grammar\n\n*One rule to ring them all*\n\nEasily build parsers / transformers for structured data.","ref":"readme.html"},{"type":"extras","title":"Overview - Grammar","doc":"An Elixir library that helps building parsers / transformers for LL(1) structured data, via two simple macros, and a protocol.","ref":"readme.html#overview"},{"type":"extras","title":"Usage - Grammar","doc":"> Please have a look at the examples for more advanced use cases\n\n`use Grammar` and write \"rule\" functions that have direct access to parsing results.\n\nYou are fully in control of what is returned by each rule function.\n\n#### Example\n\nThe following module extracts a list of integers from string of comma separated number, and enclosed by square brackets.\n\n```elixir\ndefmodule NumberListReader do\n  use Grammar\n\n  rule list(\"[\", :list_tail, \"]\") do\n    [_, list, _] = params\n    case list do\n      nil -> []\n      _ -> list\n    end\n  end\n\n  rule? list_tail(:number, :list_tail_tail) do\n    [number, list] = params\n    [number | (list || [])]\n  end\n\n  rule? list_tail_tail(\",\", :number, :list_tail_tail) do\n    [_, number, list] = params\n    [number | (list || [])]\n  end\n\n  rule number(~r/[0-9]+/) do\n    [number] = params\n    String.to_integer(number)\n  end\nend\n\n$iex> NumberListReader.parse(\"[1, 2, 6, 12]\")\n{:ok, [1, 2, 6, 12]}\n```","ref":"readme.html#usage"},{"type":"extras","title":"MIT License","doc":"# MIT License\n\nCopyright (c) 2024 Nicolas Michel\n\nPermission is hereby granted, free of charge, to any person obtaining\na copy of this software and associated documentation files (the\n\"Software\"), to deal in the Software without restriction, including\nwithout limitation the rights to use, copy, modify, merge, publish,\ndistribute, sublicense, and/or sell copies of the Software, and to\npermit persons to whom the Software is furnished to do so, subject to\nthe following conditions:\n\nThe above copyright notice and this permission notice shall be\nincluded in all copies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND,\nEXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF\nMERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND\nNONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE\nLIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION\nOF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION\nWITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.","ref":"license.html"},{"type":"extras","title":"Math expression to french","doc":"# Math expression to french\n\nThe module `MathToFrench` can read simple integer math expressions, like \"1+2\", \"(42-2)*12 / 5\",\nand rewrite them in ... french.\n\n```\niex> MathToFrench.parse(\"(42-2)*12 / 5\")\n{:ok, \"(42 moins 2) multiplié par 12 divisé par 5\"}\n```\n\nHere follows its declaration.\n\n```elixir\ndefmodule MathToFrench do\n  use Grammar\n\n  # This rule wille be the entry point.\n  # Note that we could get rid of it, as it add no real value to\n  # the output.\n  rule start(:expression) do\n    [program] = params\n    \"#{program}\"\n  end\n\n  # a math expression is a composed of \"terms\" joined by\n  # symbol + or -.\n  # This ensure precedence of other operators over the them.\n  rule expression(:term, :expression_cont) do\n    join_non_nil(params)\n  end\n\n  rule? expression_cont(\"+\", :term, :expression_cont) do\n    [_, term, cont] = params\n    join_non_nil([\"plus\", term, cont])\n  end\n\n  rule? expression_cont(\"-\", :term, :expression_cont) do\n    [_, term, cont] = params\n    join_non_nil([\"moins\", term, cont])\n  end\n\n  # A term is composed of elements joined by operators * or /.\n  # Those elements are reduced by rule `factor`.\n  rule term(:factor, :term_cont) do\n    join_non_nil(params)\n  end\n\n  rule? term_cont(\"*\", :factor, :term_cont) do\n    [_, factor, cont] = params\n    join_non_nil([\"multiplié par\", factor, cont])\n  end\n\n  rule? term_cont(\"/\", :factor, :term_cont) do\n    [_, factor, cont] = params\n    join_non_nil([\"divisé par\", factor, cont])\n  end\n\n  # Eventually a factor is either a integer number or\n  # a parenthesized expression, as in rule `expression`.\n  rule factor(:number) do\n    [number] = params\n    number\n  end\n\n  rule factor(\"(\", :expression, \")\") do\n    [_, expression, _] = params\n    \"(#{expression})\"\n  end\n\n  rule number(~r/[0-9]+/) do\n    [number] = params\n    number\n  end\n\n  defp join_non_nil(params) when is_list(params) do\n    params |> Enum.reject(&is_nil/1) |> Enum.join(\" \")\n  end\nend\n```","ref":"math_to_french.html"},{"type":"extras","title":"Math expression to function","doc":"# Math expression to function\n\n> ⚠️ Consider reading [Math to french](./math_to_french.md) example first.\n\nThe module `MathToFunction` can read simple integer math expressions too,\nbut this time the parse/1 function returns a function that produce the value of the expression when called.\n\n```\niex> {:ok, calcfn} = MathToFunction.parse(\"(42-2)*12 / 5\")\niex> calcfn.()\n96.0\n```\n\nHere follows its declaration, a little bit more chalenging.\n\n```elixir\ndefmodule MathToFunction do\n  use Grammar\n\n  rule expression(:term, :expression_cont) do\n    [term, cont] = params\n    fn -> (cont || & &1).(term.()) end\n  end\n\n  rule? expression_cont(\"+\", :term, :expression_cont) do\n    [_, term, cont] = params\n    fn left -> (cont || & &1).((left + term.())) end\n  end\n\n  rule? expression_cont(\"-\", :term, :expression_cont) do\n    [_, term, cont] = params\n    fn left -> (cont || & &1).((left - term.())) end\n  end\n\n  rule term(:factor, :term_cont) do\n    [factor, cont] = params\n    fn -> (cont || & &1).(factor.()) end\n  end\n\n  rule? term_cont(\"*\", :factor, :term_cont) do\n    [_, factor, cont] = params\n    fn left -> (cont || & &1).((left * factor.())) end\n  end\n\n  rule? term_cont(\"/\", :factor, :term_cont) do\n    [_, factor, cont] = params\n    fn left -> (cont || & &1).((left / factor.())) end\n  end\n\n  rule factor(:number) do\n    [number] = params\n    number\n  end\n\n  rule factor(\"(\", :expression, \")\") do\n    [_, expression, _] = params\n    expression\n  end\n\n  rule number(~r/[0-9]+/) do\n    [number] = params\n    number_value = String.to_integer(number)\n    fn -> number_value end\n  end\nend\n```","ref":"math_to_function.html"},{"type":"extras","title":"List of IP addresses","doc":"# List of IP addresses\n\nThis example illustrates the use of [TokenExtractor](`Grammar.Tokenizer.TokenExtractor`) protocol for user defined token.\n\nFirst, we define the token prototype to catch IP address. In our simple case, the field `ip` will contain the list of integer values that compose the address.\n\n```elixir\ndefmodule IP do\n  defstruct ip: nil\nend\n```\n\nThen we implement the [TokenExtractor](`Grammar.Tokenizer.TokenExtractor`) protocol for that new token type.\n\nHere we are providing a naive implementation, based on a (poor) regular expression and [TokenExtractorHelper](`Grammar.Tokenizer.TokenExtractorHelper`) helper module.\n\nOne is free to implement a more robust / efficient version, as long as it respects the protocol.\n\n```elixir\ndefimpl Grammar.Tokenizer.TokenExtractor, for: IP do\n  @pattern ~r/^\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}/\n\n  alias Grammar.Tokenizer.TokenExtractorHelper\n\n  def try_read(_token_prototype, input_string) do\n    TokenExtractorHelper.try_read_from_regex(@pattern, input_string)\n    |> case do\n      nil ->\n        nil\n\n      {extracted_string, length} ->\n        token =\n          extracted_string\n          |> String.split(\".\")\n          |> then(&struct(IP, ip: &1))\n\n        {token, length}\n    end\n  end\n\n  def match?(%IP{}, %IP{}), do: true\n  def match?(_prototype, _token), do: false\nend\n```\n\nThen use this new token to define rules.\n\n```elixir\ndefmodule IPListReader do\n  use Grammar\n\n  rule ips(%IP{}, :ips_tail) do\n    [ip, tail] = params\n    [ip | (tail || [])]\n  end\n\n  rule? ips_tail(\",\", %IP{}, :ips_tail) do\n    [_, ip, tail] = params\n    [ip | (tail || [])]\n  end\nend\n```\n\nAnd we are done !\n\n```elixir\n$iex> IPListReader.parse(\"\n   1.2.3.4 ,\n   3.2.3.4 ,\n   4.2.3.4\n\")\n\n{:ok,\n [\n   %IP{ip: [\"1\", \"2\", \"3\", \"4\"]},\n   %IP{ip: [\"3\", \"2\", \"3\", \"4\"]},\n   %IP{ip: [\"4\", \"2\", \"3\", \"4\"]}\n ]}\n```","ref":"ip_list.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}